/*
  PCA9685_servo_demo.ino
  Code by lingib
  https://www.instructables.com/member/lingib/instructables/
  Last update 12 January 2022

  ------
  ABOUT
  ------
  This code demonstrates five servos running at x1, x2, x3, x4, and x5 speeds
  All servo arms will be in sync every 10 arm movements of the fastest servo.

  A single PCA9685 servo module is used for controlling the servos
  https://cdn-learn.adafruit.com/downloads/pdf/16-channel-pwm-servo-driver.pdf

  ----------
  COPYRIGHT
  ----------
  This software is distributed in the hope that it will be useful,
  but without and warranty; without even the implied warranty of
  merchantability or ffitness for a particular purpose. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License. If
  not, see <http://www.gnu.org/licenses/>.
*/

//Libraries
#include <Wire.h>                       //https://www.arduino.cc/en/reference/wire
#include <Adafruit_PWMServoDriver.h>    //https://github.com/adafruit/Adafruit-PWM-Servo-Driver-Library

// ----- Constants
#define numberServos 16

// ----- Parameters
//           [servoNumber] =  0     1     2     3     4     5     6     7     8     9     10    11    12    13    14    15
int MIN_IMP [numberServos] = {500,  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,  500};   // Minimum pulse width (IMPulse)
int MAX_IMP [numberServos] = {2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500, 2500};  // Maximum pulse width
int MIN_ANG [numberServos] = {0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0};     // Angle @ MIN_IMP
int MAX_ANG [numberServos] = {180,  180,  180,  180,  180,  180,  180,  180,  180,  180,  180,  180,  180,  180,  180,  180};   //  Angle @ MAX_IMP

// ----- Objects
/*
   The default address for each PCA9685 servo module is 0x40
   Any address in the range 0x40 (B01000000) to 0x7F (B01111111) may be chosen by bridging the open jumper positions
   In theory 6 jumpers means we can attach up to 64 PCA9685 servo modules, each with 16 servos, to our arduino for a total of 1024 servos!
   But in practice, possibly due to I2C limitations, it appears that only 62 servo modules maybe chained to get a total of 992 servos.
   https://cdn-learn.adafruit.com/downloads/pdf/16-channel-pwm-servo-driver.pdf
*/
//Adafruit_PWMServoDriver pca0 = Adafruit_PWMServoDriver(0x40);   // B0100 0000 (no jumpers)
//Adafruit_PWMServoDriver pca1 = Adafruit_PWMServoDriver(0x41);   // B0100 0001 (1=jumper; 0=no jumper)
//Adafruit_PWMServoDriver pca2 = Adafruit_PWMServoDriver(0x42);   // B0100 0010
//Adafruit_PWMServoDriver pca3 = Adafruit_PWMServoDriver(0x43);   // B0100 0011
//Adafruit_PWMServoDriver pca4 = Adafruit_PWMServoDriver(0x44);   // B0100 0100
Adafruit_PWMServoDriver pca5 = Adafruit_PWMServoDriver(0x45);     // B0100 0101 (I'm using address 0x45 in this example)

// --------------
//  setup()
// --------------
void setup() {
  pca5.begin();                                                                // Initialise
  pca5.setPWMFreq(50);  // Analog servos run at ~50 Hz (20ms) updates
}

// --------------
//  loop()
// --------------
void loop() {
  servoTest();
}

// --------------
//  servoTest()
// --------------
void servoTest() { /* function pcaScenario */

  // ----- Locals
  int pulseWidth;
  int servo0 = 0;           // Servo position on module pca5
  int servo1 = 1;
  int servo2 = 2;
  int servo3 = 3;
  int servo4 = 4;

  const int step0 = 1;      // Step size in degrees
  const int step1 = 2;
  const int step2 = 3;
  const int step3 = 4;
  const int step4 = 5;

  static int angle0 = 0;    // Current angle in degrees
  static int angle1 = 0;
  static int angle2 = 0;
  static int angle3 = 0;
  static int angle4 = 0;

  static bool dir0 = true;  // Current direction
  static bool dir1 = true;
  static bool dir2 = true;
  static bool dir3 = true;
  static bool dir4 = true;

  if ((dir0) && (angle0 + step0) > 180) dir0 = false;
  if (!(dir0) && (angle0 - step0) < 0) dir0 = true;
  (dir0) ? angle0 += step0 : angle0 -= step0;
  pca5.writeMicroseconds(servo0, jointToImp(angle0, servo0));

  if ((dir1) && (angle1 + step1) > 180) dir1 = false;
  if (!(dir1) && (angle1 - step1) < 0) dir1 = true;
  (dir1) ? angle1 += step1 : angle1 -= step1;
  pca5.writeMicroseconds(servo1, jointToImp(angle1, servo1));

  if ((dir2) && (angle2 + step2) > 180) dir2 = false;
  if (!(dir2) && (angle2 - step2) < 0) dir2 = true;
  (dir2) ? angle2 += step2 : angle2 -= step2;
  pca5.writeMicroseconds(servo2, jointToImp(angle2, servo2));

  if ((dir3) && (angle3 + step3) > 180) dir3 = false;
  if (!(dir3) && (angle3 - step3) < 0) dir3 = true;
  (dir3) ? angle3 += step3 : angle3 -= step3;
  pca5.writeMicroseconds(servo3, jointToImp(angle3, servo3));

  if ((dir4) && (angle4 + step4) > 180) dir4 = false;
  if (!(dir4) && (angle4 - step4) < 0) dir4 = true;
  (dir4) ? angle4 += step4 : angle4 -= step4;
  pca5.writeMicroseconds(servo4, jointToImp(angle4, servo4));

  delay(30);
}

// --------------
//  jointToImp()
// --------------
/*
   The maths behind this function is explained here
   https://www.aranacorp.com/en/using-a-pca9685-module-with-arduino/
*/
int jointToImp(double x, int i) { /* function jointToImp */
  //Convert joint angle into pwm command value
  int imp = (x - MIN_ANG[i]) * (MAX_IMP[i] - MIN_IMP[i]) / (MAX_ANG[i] - MIN_ANG[i]) + MIN_IMP[i];
  imp = max(imp, MIN_IMP[i]);
  imp = min(imp, MAX_IMP[i]);
  return imp;
}
